TEMPLATE = """
### <a href="#{{node.anchor}}" id="{{node.anchor}}">{{node.kind.value}} {{node.name_long if node.is_group else node.name_short}} {{node.overload_suffix}}</a>

```cpp
{{node.codeblock}}
```
{% if node.has_brief -%}
{{node.brief}}
{%- endif -%}

{% if node.has_details -%}
{{node.details}}
{%- endif -%}

{% if node.reimplements %}
Implements [*{{node.reimplements.name_long}}*]({{node.reimplements.url}})
{% endif %}
"""
#"\n"



import os
import re
import traceback
from xml.etree import ElementTree
from xml.etree.ElementTree import Element as Element
from doxybook.constants import Kind, Visibility, OVERLOAD_OPERATORS
from doxybook.cache import Cache
from doxybook.xml_parser import XmlParser
from doxybook.markdown import escape
from doxybook.utils import split_safe
from doxybook.property import Property


class Node:
    _output_dir = ''
    def __init__(self, xml_file: str, xml: Element, cache: Cache, parser: XmlParser, \
            parent: 'Node', refid: str = None, position: str = None , \
            doc_type = 'all', rel_path = ''):
        self._children: ['Node'] = []
        self._cache: Cache = cache
        self._parser: XmlParser = parser
        self._parent = parent
        self._position = position
        self.doc_type = doc_type
        self.rel_path = rel_path

        if xml_file == 'root':
            self._refid = 'root'
            self._kind = Kind.from_str('root')
            self._name = 'root'
            self._xml = None

        elif xml is None:
            print('Loading XML from: ' + xml_file)
            self._dirname = os.path.dirname(xml_file)
            self._position = self._dirname
            self._xml = ElementTree.parse(xml_file).getroot().find('compounddef')
            if self._xml is None:
                raise Exception('File ' + xml_file + ' has no <compounddef>')
            self._kind = Kind.from_str(self._xml.get('kind'))
            self._refid = self._xml.get('id')
            self._name = self._xml.find('compoundname').text
            self._cache.add_onname(self._name, self)
            self._cache.add(self._refid, self)
            self._static = False

            print('Parsing: ' + self._refid)
            self._check_for_children()
            
            title = self._xml.find('title')
            if title is not None:
                self._title = title.text
            else:
                self._title = self._name

        else:
            self._xml = xml
            if self._parent.kind == Kind.ENUM:
                self._kind = Kind.from_str('enumvalue')
            else:
                self._kind = Kind.from_str(self._xml.get('kind'))
            if refid is not None:
                self._refid = refid
            else:
                self._refid = self._xml.get('id')
            self._cache.add(self._refid, self)

            self._check_attrs()
            self._title = self._name
            self._cache.add_onname(self._name, self)
            if self._kind == Kind.ENUM:
                enumvalues = self._xml.findall('enumvalue')
                for enumvalue in enumvalues:
                    if not enumvalue.get('id'):
                        continue
                    child = Node(None, enumvalue, self._cache, self._parser,
                            self, position = self._position)
                    self.add_child(child)

        self._details = Property.Details(self._xml, parser, self._kind)
        self._brief = Property.Brief(self._xml, parser, self._kind)
        self._includes = Property.Includes(self._xml, parser, self._kind)
        self._type = Property.Type(self._xml, parser, self._kind)
        self._location = Property.Location(self._xml, parser, self._kind)
        self._params = Property.Params(self._xml, parser, self._kind)
        self._templateparams = Property.TemplateParams(self._xml, parser, self._kind)
        self._specifiers = Property.Specifiers(self._xml, parser, self._kind)
        self._values = Property.Values(self._xml, parser, self._kind)
        self._initializer = Property.Initializer(self._xml, parser, self._kind)
        self._definition = Property.Definition(self._xml, parser, self._kind)
        self._programlisting = Property.Programlisting(self._xml, parser, self._kind)
        self._CodeBlock = Property.CodeBlock(self._xml, parser, self._kind)

        self._output_dir = ''
        if xml_file:
            self._output_dir = os.path.basename(os.path.dirname(xml_file))
        elif self._parent:
            self._output_dir = self._parent._output_dir
        else:
            print('No found output dir')
            pass

    #def node_replace(rep, replace, node):
    def list_node(self, node, leval = 0):
        n = leval
        print('\t'*n, node._kind, node._name, 'now')
        if node.has_children:
            n += 1
            for child in node._children:
                print('\t'*n, child._kind, child._name, 'children')
                if child.has_children:
                    self.list_node(child, n)

    def add_child(self, child: 'Node'):
        self._children.append(child)

    def sort_children(self):
        self._children.sort(key=lambda x: x._name, reverse=False)

    def _check_for_children(self):
        for innergroup in self._xml.findall('innergroup'):
            refid = innergroup.get('refid')
            if self._kind == Kind.GROUP or self._kind == Kind.DIR or self._kind == Kind.FILE:
                try:
                    child = self._cache.get(refid)
                    self.add_child(child)
                    continue
                except:
                    pass
            child = Node(os.path.join(self._dirname, refid + '.xml'), None, self._cache, self._parser, self, \
                    doc_type = self.doc_type, rel_path = self.rel_path)
            child._visibility = Visibility.PUBLIC
            self.add_child(child)

        for innerclass in self._xml.findall('innerclass'):
            refid = innerclass.get('refid')
            prot = Visibility(innerclass.get('prot'))
            if prot == Visibility.PRIVATE:
                continue
                
            if self._kind == Kind.GROUP or self._kind == Kind.DIR or self._kind == Kind.FILE:
                try:
                    child = self._cache.get(refid)
                    self.add_child(child)
                    continue
                except:
                    pass

            try:
                child = Node(os.path.join(self._dirname, refid + '.xml'), None, self._cache, self._parser, self, 
                    doc_type = self.doc_type, rel_path = self.rel_path)
            except FileNotFoundError as e:
                child = Node(os.path.join(self._dirname, refid + '.xml'), Element('compounddef'), self._cache, self._parser, self, refid=refid, \
                    doc_type = self.doc_type, rel_path = self.rel_path)
                child._name = innerclass.text
            child._visibility = prot
            self.add_child(child)

        for innerfile in self._xml.findall('innerfile'):
            refid = innerfile.get('refid')
            if self._kind == Kind.DIR:
                try:
                    child = self._cache.get(refid)
                    self.add_child(child)
                    continue
                except:
                    pass

            child = Node(os.path.join(self._dirname, refid + '.xml'), None, self._cache, self._parser, self, \
                    doc_type = self.doc_type, rel_path = self.rel_path)
            child._visibility = Visibility.PUBLIC
            self.add_child(child)

        for innerdir in self._xml.findall('innerdir'):
            refid = innerdir.get('refid')
            if self._kind == Kind.DIR:
                try:
                    child = self._cache.get(refid)
                    self.add_child(child)
                    continue
                except:
                    pass

            child = Node(os.path.join(self._dirname, refid + '.xml'), None, self._cache, self._parser, self, \
                    doc_type = self.doc_type, rel_path = self.rel_path)
            child._visibility = Visibility.PUBLIC
            self.add_child(child)

        for innernamespace in self._xml.findall('innernamespace'):
            refid = innernamespace.get('refid')

            if self._kind == Kind.GROUP or self._kind == Kind.DIR or self._kind == Kind.FILE:
                try:
                    child = self._cache.get(refid)
                    self.add_child(child)
                    continue
                except:
                    pass

            child = Node(os.path.join(self._dirname, refid + '.xml'), None, self._cache, self._parser, self, \
                    doc_type = self.doc_type, rel_path = self.rel_path)
            child._visibility = Visibility.PUBLIC
            self.add_child(child)

        for sectiondef in self._xml.findall('sectiondef'):
            for memberdef in sectiondef.findall('memberdef'):
                kind = Kind.from_str(memberdef.get('kind'))
                if kind.is_language():
                    if self._kind == Kind.GROUP or self._kind == Kind.DIR or self._kind == Kind.FILE:
                        refid = memberdef.get('id')
                        try:
                            child = self._cache.get(refid)
                            self.add_child(child)
                            continue
                        except:
                            pass
                    child = Node(None, memberdef, self._cache, self._parser,
                            self, position = self._position, \
                            doc_type = self.doc_type, rel_path = self.rel_path)
                    #if child._kind == Kind.ENUM:
                    #    print('aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa')
                    #    print(child.name)
                    #    print(child._position)
                    #    print(memberdef.get('enumvalue'))
                    self.add_child(child)

    def _check_attrs(self):
        prot = self._xml.get('prot')
        self._visibility = Visibility(prot) if prot is not None else Visibility.PUBLIC
    
        static = self._xml.get('static')
        self._static = static == 'yes'

        explicit = self._xml.get('explicit')
        self._explicit = explicit == 'yes'

        mutable = self._xml.get('mutable')
        self._mutable = mutable == 'yes'

        inline = self._xml.get('inline')
        self._inline = inline == 'yes'

        const = self._xml.get('inline')
        self._const = const == 'yes'

        name = self._xml.find('name')
        if name is not None:
            self._name = name.text
        else:
            self._name = ''

        virt = self._xml.get('virt')
        if virt:
            self._virtual = virt == 'virtual' or virt == 'pure-virtual'
            self._pure = virt == 'pure-virtual'
        else:
            self._virtual = False
            self._pure = False

    def has(self, visibility: str, kinds: [str], static: bool) -> bool:
        return len(self.query(visibility, kinds, static)) > 0

    def query(self, visibility: str, kinds: [str], static: bool) -> ['Node']:
        ret = []
        visibility = Visibility(visibility)
        kinds = list(map(lambda kind: Kind.from_str(kind), kinds))
        for child in self._children:
            if child._visibility == visibility and child._kind in kinds and child._static == static:
                ret.append(child)
        return ret

    @property
    def is_static(self) -> bool:
        return self._static

    @property
    def is_explicit(self) -> bool:
        return self._explicit

    @property
    def is_const(self) -> bool:
        return self._const

    @property
    def is_inline(self) -> bool:
        return self._inline

    @property
    def is_mutable(self) -> bool:
        return self._mutable

    @property
    def is_virtual(self) -> bool:
        return self._virtual

    @property
    def output_dir(self) -> str:
        return self._output_dir

    @property
    def is_pure(self) -> bool:
        return self._pure

    @property
    def has_children(self) -> bool:
        return len(self._children) > 0

    @property
    def children(self) -> ['Node']:
        return self._children

    @property
    def parent(self) -> 'Node':
        return self._parent

    @property
    def is_function(self) -> bool:
        return self._kind.is_function()

    @property
    def is_variable(self) -> bool:
        return self._kind.is_variable()

    @property
    def is_namespace(self) -> bool:
        return self._kind.is_namespace()

    @property
    def is_class(self) -> bool:
        return self._kind.is_class()

    @property
    def is_struct(self) -> bool:
        return self._kind.is_struct()

    @property
    def is_enum(self) -> bool:
        return self._kind.is_enum()

    @property
    def is_class_or_struct(self) -> bool:
        return self._kind.is_class_or_struct()

    @property
    def is_interface(self) -> bool:
        return self._kind.is_interface()

    @property
    def is_typedef(self) -> bool:
        return self._kind.is_typedef()

    @property
    def is_define(self) -> bool:
        return self._kind.is_define()

    @property
    def is_union(self) -> bool:
        return self._kind.is_union()

    @property
    def is_group(self) -> bool:
        return self._kind.is_group()

    @property
    def is_language(self) -> bool:
        return self._kind.is_language()

    @property
    def is_root(self) -> bool:
        return self._kind.is_root()

    @property
    def is_parent(self) -> bool:
        return self._kind.is_parent()

    @property
    def is_friend(self) -> bool:
        return self._kind.is_friend()

    @property
    def is_file(self) -> bool:
        return self._kind.is_file()

    @property
    def is_function(self) -> bool:
        return self._kind.is_function()

    @property
    def is_dir(self) -> bool:
        return self._kind.is_dir()

    @property
    def is_page(self) -> bool:
        return self._kind.is_page()

    @property
    def name(self) -> str:
        return self._name

    @property
    def title(self) -> str:
        return self._title

    @property
    def refid(self) -> str:
        return self._refid

    @property
    def kind(self) -> str:
        return self._kind

    @property
    def is_operator(self) -> bool:
        return self._name in OVERLOAD_OPERATORS

    @property
    def operators_total(self) -> int:
        total = 0
        for child in self.children:
            if child.name in OVERLOAD_OPERATORS:
                total += 1
        return total

    @property
    def operator_num(self) -> int:
        total = 0
        for child in self.parent.children:
            if child.is_function and child.name.replace(' ', '') in OVERLOAD_OPERATORS:
                total += 1
            if child.refid == self._refid:
                break
        return total

    @property
    def name_url_safe(self) -> str:
        name = self.name_tokens[-1]
        name = name.replace(' ', '-').replace('_', '-').replace('=', '').replace('~', '').lower()
        return name

    @property
    def anchor(self) -> str:
        name = ''
        if self._name.replace(' ', '') in OVERLOAD_OPERATORS:
            num = self.operator_num
            if num > 1:
                name = 'operator-' + str(self.operator_num)
            else:
                name = 'operator'
        elif self.is_overloaded:
            name = self.name_url_safe + '-' + str(self.overload_num) + '-' + str(self.overload_total)
        else:
            name = self.name_url_safe

        if name.startswith('-'):
            name = name[1:]
        return self._kind.value + '-' + name

    @property
    def url(self) -> str:
        if self.is_parent or self.is_group or self.is_file or self.is_dir or self.is_page or self.is_function:
            if self._position:
                dirs = self._position.split('/')
                rootdir = ''
                is_getxml = False
                for dir in dirs:
                    if dir == 'XML':
                        rootdir = '/'
                        is_getxml = True
                        continue
                    if is_getxml:
                        rootdir = rootdir + dir + '/'
                #/home/gx/goxceed/docs/XML/gxdownloader
                return rootdir + self._refid + '.md'
                #return '../' + os.path.basename(self._position) + '/' + self._refid + '.md'
            else:
                return  self._refid + '.md'
        elif self.is_enum:
            return  self._refid + '.md'
        else:
            return self._parent.url + '#' + self.anchor

    @property
    def url_source(self) -> str:
        if self.is_parent or self.is_group or self.is_file or self.is_dir:
            return self._refid + '_source.md'
        else:
            return self._refid + '.md'

    @property
    def filename(self) -> str:
        return self._refid + '.md'

    @property
    def root(self) -> 'Node':
        if self._kind == Kind.ROOT:
            return self
        else:
            return self._parent.root

    @property
    def name_tokens(self) -> [str]:
        if self.is_dir or self.is_file:
            return self._name.split('/')
        return split_safe(self._name, '::')
    
    @property
    def name_short(self) -> str:
        return escape(self.name_tokens[-1])

    @property
    def name_long(self) -> str:
        try:
            if self._parent.is_parent:
                return self._parent.name_long + '::' + escape(self.name_tokens[-1])
            else:
                return escape(self._name)
        except Exception as e:
            raise e

    @property
    def name_full_unescaped(self) -> str:
        if self._parent is not None and not self._parent.is_root and self._parent.is_parent:
            return self._parent.name_full_unescaped + '::' + self.name_tokens[-1]
        else:
            return self.name_tokens[-1]

    @property
    def overload_total(self) -> int:
        if self._parent is not None:
            if self._parent.is_class_or_struct:
                count = 0
                for neighbour in self._parent.children:
                    if neighbour.name == self.name:
                        count += 1
                return count
        return 0

    @property
    def overload_num(self) -> int:
        if self._parent is not None:
            if self._parent.is_class_or_struct:
                count = 0
                for neighbour in self._parent.children:
                    if neighbour.name == self.name:
                        count += 1
                    if neighbour.refid == self.refid:
                        break
                return count
        return 0

    @property
    def is_overloaded(self) -> bool:
        return self.overload_total > 1

    @property
    def overload_suffix(self) -> str:
        if self.is_operator:
            return ''
        
        total = self.overload_total
        if total > 1:
            return '[' + str(self.overload_num) + '/' + str(total) + ']'
        else:
            return ''

    @property
    def parents(self) -> ['Node']:
        if self._refid == 'dir_e76b55871a3ceedcbe1f9359e5753575':
            print(id(self))
            print('debugger')
        ret = []
        if self._parent is not None:
            if self._parent.is_language or self._parent.is_dir:
                ret.extend(self.parent.parents)
        ret.append(self)
        return ret

    @property
    def suffix(self) -> str:
        if self.is_parent:
            if self._templateparams.has():
                return '&lt;' + ', '.join(self._templateparams.array(notype=True)) + '&gt;'
            else:
                return ''
        elif self.is_function:
            return self._specifiers.md()
        elif self.is_variable:
            if self._initializer.has():
                return ' = ' + self._initializer.md()
            else:
                return ''
        elif self.is_define:
            params = self._params.md()
            test = self._initializer.md()
            # Do not use variable_initializer if it is
            if '\n' in test:
                return ' (' + params + ')'
            return '(' + params + ') ' + test
        else:
            return ''

    @property
    def prefix(self) -> str:
        if self.is_function:
            ret = []
            if self.is_virtual:
                ret.append('virtual')
            return ' '.join(ret)
        elif self.kind is Kind.VARIABLE:
            return ''
        else:
            return self.kind.value
    
    @property
    def codeblock(self) -> str:
        code = []
        if self.is_function or self.is_friend:
            if self._templateparams.has():
                code.append('template<' + self._templateparams.plain() + '>')

            typ = self._type.plain()
            if typ:
                typ += ' '
            if self.is_virtual:
                typ = 'virtual ' + typ
            if self.is_explicit:
                typ = 'explicit ' + typ
            if self.is_inline:
                typ = 'inline ' + typ
            if self.is_static:
                typ = 'static ' + typ

            if self._params.has():
                code.append(typ + self.name_full_unescaped + ' (')
                params = self._params.array(plain=True)
                for i, param in enumerate(params):
                    if i + 1 >= len(params):
                        code.append('    ' + param)
                    else:
                        code.append('    ' + param + ',')
                code.append(') ' + self._specifiers.plain())
            else:
                code.append(typ + self.name_full_unescaped + ' () ' + self._specifiers.plain())

        elif self.is_enum:
            if self._values.has():
                code.append('enum ' + self.name_full_unescaped + ' {')
                values = self._values.array(plain=True)
                for i, value in enumerate(values):
                    value = re.sub('\*\*', '', value)
                    value = re.sub('\\\\', '', value)
                    if i + 1 >= len(values):
                        code.append('    ' + value)
                    else:
                        code.append('    ' + value + ',')
                code.append('};')
            else:
                code.append('enum ' + self.name_full_unescaped + ';')

        elif self.is_define:
            if self._params.has():
                code.append('#define ' + self.name_full_unescaped + ' (')
                params = self._params.array(plain=True)
                for i, param in enumerate(params):
                    if i + 1 >= len(params):
                        code.append('    ' + param)
                    else:
                        code.append('    ' + param + ',')
                code.append(') ' + self._initializer.plain())
            else:
                code.append('#define ' + self.name_full_unescaped + ' () ' + self._specifiers.plain())
        elif self.is_struct or self.is_union:
            pass
            if self.is_struct:
                code.append('struct ' + self.name_full_unescaped + ' {')
            else:
                name = self.name_full_unescaped
                name = re.sub('.+::','',name)
                code.append('union ' + name + ' {')
            if self.has_children:
                definitions = []
                for child in self._children:
                    if child.name and child.type:
                        definitions.append(child._definition.plain())
                        typ = child._type.plain();
                        unioncode = ''
                        if 'union' in typ:
                            if '::' in typ:
                                typ = re.sub('.+::','',typ)
                                if typ[0] == '@':
                                    typ = ''
                            node = None
                            if not '@' in child._type.plain():
                                typ = 'union ' + typ
                                nodename = re.sub('union ','',child._type.plain())
                                node = self._cache.get_on_name(nodename)
                                unioncode = re.sub('union.+{', '{', node.codeblock)
                            else:
                                nodename = child.name
                                node = self._cache.get_on_name(nodename)
                                typ = 'union'
                                #unioncode = 'union '
                            unioncode = re.sub('\n','\n    ', unioncode)
                        name = child.name;
                        if unioncode:
                            code.append('    ' + typ + unioncode + name  + ';')
                        else:
                            code.append('    ' + typ + ' ' + name + ';')
                for child in self._children:
                    if child.name and ( not child.type):
                        get_union = 0
                        for definition1 in definitions:
                            if 'union ' + child.name in definition1:
                                continue
                        get_union = 1
                        if get_union:
                            continue
                        unioncode = child.codeblock
                        if 'union' in unioncode:
                            unioncode = re.sub(' .+::', ' ', unioncode)
                            unioncode = '    ' +  unioncode + ';'
                            unioncode = re.sub('\n', '\n    ', unioncode)
                            code.append(unioncode)
                code.append('}')
        else:
            code.append(self._definition.plain())
        return '\n'.join(code)

    @property
    def has_base_classes(self) -> bool:
        return len(self._xml.findall('basecompoundref')) > 0

    @property
    def has_derived_classes(self) -> bool:
        return len(self._xml.findall('derivedcompoundref')) > 0

    @property
    def base_classes(self) -> ['Node']:
        ret = []
        for basecompoundref in self._xml.findall('basecompoundref'):
            refid = basecompoundref.get('refid')
            if refid is None:
                ret.append(basecompoundref.text)
            else:
                ret.append(self._cache.get(refid))
        return ret

    @property
    def derived_classes(self) -> ['Node']:
        ret = []
        for derivedcompoundref in self._xml.findall('derivedcompoundref'):
            refid = derivedcompoundref.get('refid')
            if refid is None:
                ret.append(derivedcompoundref.text)
            else:
                ret.append(self._cache.get(refid))
        return ret

    @property
    def has_details(self) -> bool:
        return self._details.has()

    @property
    def details(self) -> str:
        details = self._details.md()

        if re.findall('gxdocref \[', details):
            details = re.sub('gxdocref \[','[',details)
            return details
       # name = self._name.replace('_', '\\_')
        #refs = re.findall('gxdocref +[\S]+', details)
        refs = re.findall('gxdocref +[^ \n\r\t\v\f、。，；,.;]+', details)
        if refs:
            for ref in refs:
                refname = re.sub('gxdocref[\s]+', '', ref)
                refname = refname.replace('\\','')
                node = self._cache.get_on_name(refname)
                if not node:
                    continue
                refname = refname.replace('_','\\\\_')
                #details = re.sub('gxdocref *'+refname,'[**'+refname+'**]('+
                #        node.url + ')', details)
                if self.doc_type == 'all':
                    details = re.sub('gxdocref *'+refname,'[**'+refname+'**]('+
                            node.url + ')', details)
                else:
                #elif self.doc_type == 'part':
                    details = re.sub('gxdocref *'+refname,'[**'+refname+'**](/'+ \
                            self.rel_path + '/book' + node.url + ')', details)
        return details

    @property
    def has_brief(self) -> bool:
        return self._brief.has()
    
    @property
    def brief(self) -> str:
        brief = self._brief.md()
        if re.findall('gxdocref_', brief):
            brief = re.sub('gxdocref_', 'gxdocref', brief)
            brief = re.sub('(gxdocref) +\[_\*\*(.+)\*\*_\] *\(.+\) _','\g<1> \g<2> ',brief)
        #refs = re.findall('gxdocref +[\S]+', brief)
        refs = re.findall('gxdocref +[^ \n\r\t\v\f、。，；,.;]+', brief)
        if refs:
            for ref in refs:
                refname = re.sub('gxdocref[\s]+', '', ref)
                refname = refname.replace('\\','')
                if refname[len(refname)-1] == '_':
                    refname = refname[0:-1]
                node = self._cache.get_on_name(refname)
                if not node:
                    continue
                refname = refname.replace('_','\\\\_')
                if self.doc_type == 'all':
                    brief = re.sub('gxdocref *'+refname,'[**'+refname+'**]('+
                            node.url + ')', brief)
                else:
                #elif self.doc_type == 'part':
                    #tmp = node.url.split('/')
                    #tmp.insert(-1, 'book')
                    brief = re.sub('gxdocref *'+refname,'[**'+refname+'**](/' + \
                            self.rel_path + '/book' + node.url + ')', brief)
        return brief

    @property
    def has_includes(self) -> bool:
        return self._includes.has()

    @property
    def includes(self) -> str:
        return self._includes.plain()

    @property
    def has_type(self) -> bool:
        return self._type.has()

    @property
    def type(self) -> str:
        return self._type.md()

    @property
    def has_location(self) -> bool:
        return self._location.has()

    @property
    def location(self) -> str:
        return self._location.plain()

    @property
    def has_params(self) -> bool:
        return self._params.has()

    @property
    def params(self) -> str:
        if self._params.has():
            return '(' + self._params.md() + ')'
        elif self.is_function:
            return '()'
        else:
            return ''

    @property
    def has_templateparams(self) -> bool:
        return self._templateparams.has()

    @property
    def templateparams(self) -> str:
        return self._templateparams.md()

    @property
    def has_specifiers(self) -> bool:
        return self._specifiers.has()

    @property
    def specifiders(self) -> str:
        return self._specifiers.plain()

    @property
    def has_values(self) -> bool:
        return self._values.has()

    @property
    def values(self) -> str:
        return self._values.md()

    @property
    def has_initializer(self) -> bool:
        return self._initializer.has()

    @property
    def initializer(self) -> str:
        return self._initializer.md()

    @property
    def has_definition(self) -> bool:
        return self._definition.has()

    @property
    def definition(self) -> str:
        return self._definition.plain()

    @property
    def has_programlisting(self) -> bool:
        return self._programlisting.has()

    @property
    def programlisting(self) -> str:
        return self._programlisting.md()

    @property
    def is_resolved(self) -> bool:
        return True

    @property
    def reimplements(self) -> 'Node':
        reimp = self._xml.find('reimplements')
        if reimp is not None:
            return self._cache.get(reimp.get('refid'))
        else:
            return None

class DummyNode:
    def __init__(self, name_long: str, derived_classes: [Node], kind: Kind):
        self.name_long = name_long
        self.derived_classes = derived_classes
        self.kind = kind

    @property
    def is_resolved(self) -> bool:
        return False



TEMPLATE = """
# Macro List

Here are the macros with brief descriptions:


{% for node in nodes recursive %}
{%- if node.is_define %}

* **{{node.kind.value}}** [**{{node.name_short}}**]({{node.refid+'.md'}}) {{node.brief}}

{%- endif %}
{%- if node.has_children %}
  {{- loop(node.children) }}
{%- endif -%}
{%- endfor %}
"""
#{%- if node.is_define and node.brief %}



import os
import re
import string
import traceback
from typing import TextIO
from jinja2 import Template
from jinja2.exceptions import TemplateSyntaxError, TemplateError
from jinja2 import StrictUndefined
from doxybook.node import Node, DummyNode
from doxybook.constants import Kind
from doxybook.templates.annotated import TEMPLATE as ANNOTATED_TEMPLATE
from doxybook.templates.functionlist import TEMPLATE as FUNCTIONLIST_TEMPLATE
from doxybook.templates.definelist import TEMPLATE as DEFINELIST_TEMPLATE
from doxybook.templates.enumlist import TEMPLATE as ENUMLIST_TEMPLATE
from doxybook.templates.function import TEMPLATE as FUNCTION_TEMPLATE
from doxybook.templates.enum import TEMPLATE as ENUM_TEMPLATE
from doxybook.templates.member import TEMPLATE as MEMBER_TEMPLATE
from doxybook.templates.member_definition import TEMPLATE as MEMBER_DEFINITION_TEMPLATE
from doxybook.templates.member_table import TEMPLATE as MEMBER_TABLE_TEMPLATE
from doxybook.templates.namespaces import TEMPLATE as NAMESPACES_TEMPLATE
from doxybook.templates.classes import TEMPLATE as CLASSES_TEMPLATE
from doxybook.templates.hierarchy import TEMPLATE as HIEARARCHY_TEMPLATE
from doxybook.templates.index import TEMPLATE as INDEX_TEMPLATE
from doxybook.templates.modules import TEMPLATE as MODULES_TEMPLATE
from doxybook.templates.files import TEMPLATE as FILES_TEMPLATE
from doxybook.templates.programlisting import TEMPLATE as PROGRAMLISTING_TEMPLATE
from doxybook.templates.page import TEMPLATE as PAGE_TEMPLATE
from doxybook.templates.pages import TEMPLATE as PAGES_TEMPLATE
import sys

LETTERS = string.ascii_lowercase + '~_@'

funcname = sys._getframe().f_back.f_code.co_name 
linenumber = sys._getframe().f_back.f_lineno 

ADDITIONAL_FILES = {
    'Namespace List': 'namespaces.md',
    'Namespace Members': 'namespace_members.md',
    'Namespace Member Functions': 'namespace_member_functions.md',
    'Namespace Member Variables': 'namespace_member_variables.md',
    'Namespace Member Typedefs': 'namespace_member_typedefs.md',
    'Namespace Member Enumerations': 'namespace_member_enums.md',
    'Class Index': 'classes.md',
    'Class Hierarchy': 'hierarchy.md',
    'Class Members': 'class_members.md',
    'Class Member Functions': 'class_member_functions.md',
    'Class Member Variables': 'class_member_variables.md',
    'Class Member Typedefs': 'class_member_typedefs.md',
    'Class Member Enumerations': 'class_member_enums.md',
}

def generate_link(name, url) -> str:
    '''
    in modify summary
    '''
    return '* [' + name + '](' + url + ')\n'

class Generator:
    def __init__(self, target: str, ignore_errors: bool = False, allgenerator: str = None):
        self.target = target
        self.allgenerator = allgenerator

        on_undefined_class = None
        if not ignore_errors:
            on_undefined_class = StrictUndefined

        try: 
            self.annotated_template = Template(ANNOTATED_TEMPLATE, undefined=on_undefined_class)
            self.functionlist_template = Template(FUNCTIONLIST_TEMPLATE, undefined=on_undefined_class)
            self.definelist_template = Template(DEFINELIST_TEMPLATE, undefined=on_undefined_class)
            self.function_template = Template(FUNCTION_TEMPLATE, undefined=on_undefined_class)
            self.enumlist_template = Template(ENUMLIST_TEMPLATE, undefined=on_undefined_class)
            self.enum_template = Template(ENUM_TEMPLATE, undefined=on_undefined_class)
            self.member_template = Template(MEMBER_TEMPLATE, undefined=on_undefined_class)
            self.member_definition_template = Template(MEMBER_DEFINITION_TEMPLATE, undefined=on_undefined_class)
            self.member_table_template = Template(MEMBER_TABLE_TEMPLATE, undefined=on_undefined_class)
            self.namespaces_template = Template(NAMESPACES_TEMPLATE, undefined=on_undefined_class)
            self.classes_template = Template(CLASSES_TEMPLATE, undefined=on_undefined_class)
            self.hiearchy_template = Template(HIEARARCHY_TEMPLATE, undefined=on_undefined_class)
            self.index_template = Template(INDEX_TEMPLATE, undefined=on_undefined_class)
            self.modules_template = Template(MODULES_TEMPLATE, undefined=on_undefined_class)
            self.files_template = Template(FILES_TEMPLATE, undefined=on_undefined_class)
            self.programlisting_template = Template(PROGRAMLISTING_TEMPLATE, undefined=on_undefined_class)
            self.page_template = Template(PAGE_TEMPLATE, undefined=on_undefined_class)
            self.pages_template = Template(PAGES_TEMPLATE, undefined=on_undefined_class)
        except TemplateSyntaxError as e:
            raise Exception(str(e) + ' at line: ' + str(e.lineno))

    def _render(self, tmpl: Template, path: str, data: dict) -> str:
        try: 
            print('Generating', path)
            output = tmpl.render(data)

            if (not os.path.exists(os.path.dirname(path))):
                os.makedirs(os.path.dirname(path))
            with open(path, 'w+', encoding='utf-8') as file:
                file.write(output) 
        except TemplateError as e:
            raise Exception(str(e))

    def _recursive_find(self, nodes: [Node], kind: Kind):
        ret = []
        for node in nodes:
            if node.kind == kind:
                ret.append(node)
            if node.kind.is_parent():
                ret.extend(self._recursive_find(node.children, kind))
        return ret

    def _recursive_find_with_parent(self, nodes: [Node], kinds: [Kind], parent_kinds: [Kind]):
        ret = []
        for node in nodes:
            if node.kind in kinds and node.parent is not None and node.parent.kind in parent_kinds:
                ret.append(node)
            if node.kind.is_parent() or node.kind.is_dir() or node.kind.is_file():
                ret.extend(self._recursive_find_with_parent(node.children, kinds, parent_kinds))
        return ret

    def annotated(self, output_dir: str, nodes: [Node]):
        path = os.path.join(output_dir, 'annotated.md')

        data = {
            'nodes': nodes,
            'target': self.target
        }
        self._render(self.annotated_template, path, data) 

    def functionlist(self, output_dir: str, nodes: [Node]):
        path = os.path.join(output_dir, 'functionlist.md')

        data = {
            'nodes': nodes,
            'target': self.target
        }
        self._render(self.functionlist_template, path, data) 

    def definelist(self, output_dir: str, nodes: [Node]):
        path = os.path.join(output_dir, 'definelist.md')

        data = {
            'nodes': nodes,
            'target': self.target
        }
        self._render(self.definelist_template, path, data) 

    def list_node(self, node, leval = 0):
        n = leval
        print('\t'*n, node.kind, node.name, 'now')
        if node.has_children:
            n += 1
            for child in node.children:
                print('\t'*n, child.kind, child.name, 'children')
                if child.has_children:
                    self.list_node(child, n)

    def enumlist(self, output_dir: str, nodes: [Node]):
        path = os.path.join(output_dir, 'enumlist.md')

        data = {
            'nodes': nodes,
            'target': self.target
        }
        self._render(self.enumlist_template, path, data) 

    def programlisting(self, output_dir: str, node: [Node]):
        path = os.path.join(output_dir, node.refid + '_source.md')

        data = {
            'node': node,
            'target': self.target
        }
        self._render(self.programlisting_template, path, data) 

    def fileindex(self, output_dir: str, nodes: [Node]):
        path = os.path.join(output_dir, 'files.md')

        data = {
            'nodes': nodes,
            'target': self.target
        }
        self._render(self.files_template, path, data) 

    def namespaces(self, output_dir: str, nodes: [Node]):
        path = os.path.join(output_dir, 'namespaces.md')

        data = {
            'nodes': nodes,
            'target': self.target
        }
        self._render(self.namespaces_template, path, data)

    def page(self, output_dir: str, node: Node):
        path = os.path.join(output_dir, node.name + '.md')

        data = {
            'node': node,
            'target': self.target
        }
        self._render(self.page_template, path, data)

    def pages(self, output_dir: str, nodes: [Node]):
        for node in nodes:
            self.page(output_dir, node)

    def relatedpages(self, output_dir: str, nodes: [Node]):
        path = os.path.join(output_dir, 'pages.md')

        data = {
            'nodes': nodes,
            'target': self.target
        }
        self._render(self.pages_template, path, data)

    def classes(self, output_dir: str, nodes: [Node]):
        path = os.path.join(output_dir, 'classes.md')

        classes = self._recursive_find(nodes, Kind.CLASS)
        classes.extend(self._recursive_find(nodes, Kind.STRUCT))
        classes.extend(self._recursive_find(nodes, Kind.INTERFACE))
        dictionary = {}

        for letter in LETTERS:
            dictionary[letter] = []

        for klass in classes:
            if klass.name_short[0] == '\\':
                dictionary[klass.name_short[1].lower()].append(klass)
            else:
                dictionary[klass.name_short[0].lower()].append(klass)


        for letter in list(dictionary):
            if len(dictionary[letter]) == 0:
                del dictionary[letter]

        data = {
            'dictionary': dictionary,
            'target': self.target
        }
        self._render(self.classes_template, path, data) 

    def _find_base_classes(self, nodes: [Node], derived: Node):
        ret = []
        for node in nodes:
            if isinstance(node, str):
                ret.append({
                    'refid': node, 
                    'derived': derived
                })
            elif node.kind.is_parent() and not node.kind.is_namespace():
                bases = node.base_classes
                if len(bases) == 0:
                    ret.append(node)
                else:
                    ret.extend(self._find_base_classes(bases, node))
        return ret

    def modules(self, output_dir: str, nodes: [Node]):
        path = os.path.join(output_dir, 'modules.md')

        data = {
            'nodes': nodes,
            'target': self.target
        }
        self._render(self.modules_template, path, data) 

    def hierarchy(self, output_dir: str, nodes: [Node]):
        path = os.path.join(output_dir, 'hierarchy.md')

        classes = self._recursive_find(nodes, Kind.CLASS)
        classes.extend(self._recursive_find(nodes, Kind.STRUCT))
        classes.extend(self._recursive_find(nodes, Kind.INTERFACE))

        bases = self._find_base_classes(classes, None)
        deduplicated = {}

        for base in bases:
            if not isinstance(base, dict):
                deduplicated[base.refid] = base
                
        for base in bases:
            if isinstance(base, dict):
                if base['refid'] not in deduplicated:
                    deduplicated[base['refid']] = []
                deduplicated[base['refid']].append(base)

        deduplicated_arr = []
        for key, children in deduplicated.items():
            if isinstance(children, list):
                deduplicated_arr.append(DummyNode(
                    key,
                    list(map(lambda x: x['derived'], children)),
                    Kind.CLASS
                ))
            else:
                found: Node = None
                for klass in classes:
                    if klass.refid == key:
                        found = klass
                        break
                if found:
                    deduplicated_arr.append(found)

        data = {
            'classes': deduplicated_arr,
            'target': self.target
        }
        self._render(self.hiearchy_template, path, data) 

    def member(self, output_dir: str, node: Node):
        path = os.path.join(output_dir, node.filename)

        data = {
            'node': node,
            'target': self.target,
            'member_definition_template': self.member_definition_template,
            'member_table_template': self.member_table_template
        }
        self._render(self.member_template, path, data)

        if node.is_language or node.is_group or node.is_file or node.is_dir:
            self.members(output_dir, node.children)

    def file(self, output_dir: str, node: Node):
        path = os.path.join(output_dir, node.filename)

        data = {
            'node': node,
            'target': self.target,
            'member_definition_template': self.member_definition_template,
            'member_table_template': self.member_table_template
        }
        self._render(self.member_template, path, data)

        if node.is_file and node.has_programlisting:
            self.programlisting(output_dir, node)

        if node.is_file or node.is_dir:
            self.files(output_dir, node.children)

    def members(self, output_dir: str, nodes: [Node]):
        for node in nodes:
            if node.is_parent or node.is_group or node.is_file or node.is_dir:
                self.member(output_dir, node)

    def function(self, output_dir: str, node: Node):
        path = os.path.join(output_dir, node.refid + '.md')
        data = {
            'node': node,
            'target': self.target,
            'member_definition_template': self.member_definition_template,
            'member_table_template': self.member_table_template
        }
        
        self._render(self.function_template, path, data)

        if node.is_language or node.is_group or node.is_file or node.is_dir:
            self.functions(output_dir, node.children)


    def functions(self, output_dir: str, nodes: [Node]):
        for node in nodes:
            if node.is_parent or node.is_group or node.is_file or node.is_dir:
                if node.has_children:
                    for child in node.children:
                        if child.kind.is_function() and (child.brief or child.details):
                            self.function(output_dir, child)
            elif node.kind.is_function():
                self.function(output_dir, child)
            else:
                pass

    def enum(self, output_dir: str, node: Node):
        path = os.path.join(output_dir, node.refid + '.md')
        data = {
            'node': node,
            'target': self.target,
            'member_definition_template': self.member_definition_template,
            'member_table_template': self.member_table_template
        }
        self._render(self.enum_template, path, data)

        if node.is_language or node.is_group or node.is_file or node.is_dir:
            self.enums(output_dir, node.children)


    def enums(self, output_dir: str, nodes: [Node]):
        for node in nodes:
            if node.is_parent or node.is_group or node.is_file or node.is_dir:
                if node.has_children:
                    for child in node.children:
                        if child.kind.is_enum() and (child.brief or child.details):
                            self.enum(output_dir, child)
            elif node.kind.is_enum():
                self.enum(output_dir, child)
            else:
                pass
                #print('functions:', node.kind)
                        #else:
                        #    print(child.kind)
                        #    self.functions(output_dir, child)


    def files(self, output_dir: str, nodes: [Node]):
        for node in nodes:
            if node.is_file or node.is_dir:
                self.file(output_dir, node)

    def index(self, output_dir: str, nodes: [Node], kind_filters: Kind, kind_parents: [Kind], title: str):
        path = os.path.join(output_dir, title.lower().replace(' ', '_') + '.md')

        found_nodes = self._recursive_find_with_parent(nodes, kind_filters, kind_parents)
        dictionary = {}

        # Populate initial dictionary
        for letter in LETTERS:
            dictionary[letter] = []

        # Sort items into the dictionary
        for found in found_nodes:
            dictionary[found.name_tokens[-1][0].lower()].append(found)

        # Delete unused letters
        for letter in list(dictionary):
            if len(dictionary[letter]) == 0:
                del dictionary[letter]

        # Sort items if they have the same name
        sorted_dictionary = {}
        for letter, items in dictionary.items():
            d = {}
            for item in items:
                # The name of the item is not yet in the dictionary
                if item.name_short not in d:
                    d[item.name_short] = [item.parent]

                # If the key is already in the dictionary,
                # make sure there are no duplicates.
                # For example an overloaded constructor or function!
                # Only allow distinct parents
                else:
                    found = False
                    for test in d[item.name_short]:
                        if test.refid == item.parent.refid:
                            found = True
                            break
                    if not found:
                        d[item.name_short].append(item.parent)

            sorted_dictionary[letter] = d

        data = {
            'title': title,
            'dictionary': sorted_dictionary,
            'target': self.target
        }
        self._render(self.index_template, path, data)

    def _generate_recursive(self, f: TextIO, node: Node, level: int, diff: str):
        if node.kind.is_parent():
            f.write(' ' * level + generate_link(node.kind.value + ' ' + node.name, diff + '/' + node.refid + '.md'))
            for child in node.children:
                self._generate_recursive(f, child, level + 2, diff)

    def _generate_recursive_function(self, f: TextIO, node: Node, level: int, diff: str):
        if node.kind.is_function() and (node.brief or node.details):
            f.write(' ' * level + generate_link(node.kind.value + ' ' + node.name, diff + '/' + node.refid + '.md'))
        if node.kind.is_group():
            for child in node.children:
                self._generate_recursive_function(f, child, level + 2, diff)

    def _generate_recursive_enum(self, f: TextIO, node: Node, level: int, diff: str):
        #if node.kind.is_enum() and node.has_brief:
        if node.kind.is_enum():
            f.write(' ' * level + generate_link(node.kind.value + ' ' + node.name, diff + '/' + node.refid + '.md'))
        if node.kind.is_group():
            for child in node.children:
                self._generate_recursive_enum(f, child, level + 2, diff)

    def _generate_recursive_files(self, f: TextIO, node: Node, level: int, diff: str):
        if node.kind.is_file() or node.kind.is_dir():
            f.write(' ' * level + generate_link(node.name, diff + '/' + node.refid + '.md'))
            if node.kind.is_file():
                f.write(' ' * level + generate_link(node.name + ' source', diff + '/' + node.refid + '_source.md'))
            for child in node.children:
                self._generate_recursive_files(f, child, level + 2, diff)

    def _generate_recursive_groups(self, f: TextIO, node: Node, level: int, diff: str):
        if node.kind.is_group():
            f.write(' ' * level + generate_link(node.title, diff + '/' + node.refid + '.md'))
            for child in node.children:
                self._generate_recursive_groups(f, child, level + 2, diff)     

    def _generate_recursive_pages(self, f: TextIO, node: Node, level: int, diff: str):
        if node.kind.is_page():
            f.write(' ' * level + generate_link(node.title, diff + '/' + node.refid + '.md'))
            for child in node.children:
                self._generate_recursive_pages(f, child, level + 2, diff)     

    def summary(self, output_dir: str, summary_file: str, nodes: [Node], modules: [Node], files: [Node], pages: [Node], no_example = False):
        print('Modifying', summary_file)
        summaryDir = os.path.dirname(os.path.abspath(summary_file))
        output_path = os.path.abspath(output_dir)
        diff = output_path[len(summaryDir)+1:].replace('\\', '/')
        #link = os.path.basename(diff) + '/index.md'
        link = 'index.md'
        if diff:
            link = diff + '/index.md'

        content = []
        with open(summary_file, 'r') as f:
            content = f.readlines()

        start = None
        offset = 0
        end = None
        for i in range(0, len(content)):
            line = content[i]
            if start is None and re.search(re.escape(link), line):
                m = re.search('\\* \\[', line)
                if m is not None:
                    start = m.start()
                    start = i
                continue
            
            if start is not None and end is None:
                if not line.startswith(' ' * (offset + 2)):
                    end = i

        if start is None:
            print('WARNING: Could not generate summary! Unable to find \"* [...](' + link + ')\" in SUMMARY.md')
            return

        if end is None:
            end = len(content)

        with open(summary_file, 'w+') as f:
            # Write first part of the file
            for i in range(0, start+1):
                f.write(content[i])

            #f.write(' ' * (offset+2) + generate_link('Related Pages', diff + '/' + 'pages.md'))
            #for node in pages:
            #    self._generate_recursive_pages(f, node, offset + 4, diff)

            #f.write(' ' * (offset+2) + generate_link('Modules', diff + '/' + 'modules.md'))
            #for node in modules:
            #    self._generate_recursive_groups(f, node, offset + 4, diff)

            line_content = content[start]
            while content[start][offset] == ' ':
                offset = offset + 1
            if offset == 0:
                while content[start][offset] == '\t':
                    offset = offset + 8
            if start+1 < len(content):
                for i in range(start+1, len(content)):
                    is_get = 0
                    line_content = content[i]
                    keywords = ('\[数据结构\]','\[宏定义\]','\[枚举\]', '\[结构体\]', '\[函数\]', '\[Example\]','\[define','\[union','\[enum','\[struct','\[function')
                    if not line_content:
                        continue
                    for keyword in keywords:
                        if not re.findall(keyword, line_content):
                            is_get = is_get + 1
                    if is_get == len(keywords):
                        end = i
                        break

            f.write(' ' * (offset+2) + generate_link('数据结构', diff + '/' + 'annotated.md'))
            f.write(' ' * (offset+4) + generate_link('宏定义', diff + '/' + 'definelist.md'))
            f.write(' ' * (offset+4) + generate_link('枚举', diff + '/' + 'enumlist.md'))
            for node in modules:
                if self.allgenerator != 'no':
                    self._generate_recursive_enum(f, node, offset + 6, diff)  
            f.write(' ' * (offset+4) + generate_link('结构体', diff + '/' + 'annotated.md'))
            for node in nodes:
                if self.allgenerator != 'no':
                    self._generate_recursive(f, node, offset + 6, diff)  

            f.write(' ' * (offset+2) + generate_link('函数', diff + '/' + 'functionlist.md'))
            for node in modules:
                if self.allgenerator != 'no':
                    self._generate_recursive_function(f, node, offset + 4, diff)  


            if no_example == False:
                f.write(' ' * (offset+2) + generate_link('Example', diff + '/' + 'example.md'))
            #for key, val in ADDITIONAL_FILES.items():
            #    f.write(' ' * (offset+2) + generate_link(key, diff + '/' + val))

            #f.write(' ' * (offset+2) + generate_link('Files', diff + '/' + 'files.md'))
            #for node in files:
            #    self._generate_recursive_files(f, node, offset + 4, diff)    

            #f.write(' ' * (offset+2) + generate_link('File Variables', diff + '/' + 'variables.md'))
            #f.write(' ' * (offset+2) + generate_link('File Functions', diff + '/' + 'functions.md'))
            #f.write(' ' * (offset+2) + generate_link('File Macros', diff + '/' + 'macros.md'))

            ## Write second part of the file
            for i in range(end, len(content)):
                f.write(content[i])



TEMPLATE = """
# File List

Here is a list of all files with brief descriptions:

{% for node in nodes recursive %}
{%- if node.is_dir or node.is_file %}
* **{{node.kind.value}}** [**{{node.name_short}}**]({{node.url}}) {{node.brief}}
{%- if node.has_children %}
  {{- loop(node.children)|indent(2, true) }}
{%- endif %}
{%- endif -%}
{%- endfor %}
"""



TEMPLATE = """
# Class Index

{% for letter, children in dictionary.items() %}
## {{letter}}

{% for node in children -%}
* [**{{node.name_short}}**]({{node.url}})
{%- if node.parent.is_language %}
 ([**{{node.parent.name_long}}**]({{node.parent.url}}))
{%- endif -%}
{% endfor %}
{% endfor %}
"""



from xml.etree.ElementTree import Element as Element
from doxybook.markdown import Md, MdRenderer, MdParagraph, MdTable, Code, MdTableRow, MdCodeBlock, MdTableCell, MdHeader, MdImage, MdList, MdBlockQuote, MdLink, MdBold, MdItalic, MdHint, Text, Br
from doxybook.cache import Cache
from doxybook.utils import lookahead


SIMPLE_SECTIONS = {
    'see': 'See also:',
    'note': 'Note:',
    'bug': 'Bug:',
    'warning': 'Warning:',
    'return': 'Returns:',
    'returns': 'Returns:',
    'param': 'Parameters:',
    'templateparam': 'Template parameters:',
    'retval': 'Return value:',
    'author': 'Author:',
    'authors': 'Authors:',
    'since': 'Since:',
    'pre': 'Precondition:',
    'remark': 'Remark:',
    'copyright': 'Copyright:',
    'post': 'Postcondition:',
    'rcs': 'Rcs:',
    'attention': 'Attention:',
    'invariant': 'Invariant:',
    'exception': 'Exception:',
    'date': 'Date:'
}

SIMPLE_SECTIONS_HINTS_VUEPRESS = {
    'note': 'tip',
    'bug': 'danger',
    'warning': 'warning'
}

class XmlParser:
    def __init__(self, cache: Cache, target: str = 'gitbook', hints: bool = True):
        self.target = target
        self.cache = cache
        self.hints = hints

    def anchor(self, name: str) -> str:
        return '<a name=\"' + name + '\"></a>'

    def paras_as_str(self, p: Element, italic: bool = False, plain: bool = False) -> str:
        if plain:
            return self.plain_as_str(p)
        else:
            renderer = MdRenderer()
            for m in self.paras(p, italic=italic):
                m.render(renderer, '')
            return renderer.output

    def reference_as_str(self, p: Element) -> str:
        renderer = MdRenderer()
        refid = p.get('refid')
        if refid is not None:
            m = MdLink([MdBold([Text(p.text)])], refid)
            m.render(renderer, '')
            return renderer.output
        else:
            return p.text

    def programlisting_as_str(self, p: Element) -> str:
        renderer = MdRenderer()
        for m in self.programlisting(p):
            m.render(renderer, '')
        return renderer.output

    def plain_as_str(self, p: Element) -> str:
        return ' '.join(self.plain(p)).strip()

    def plain(self, p: Element) -> [str]:
        ret = []
        if p is None:
            return ret
        if p.text:
            ret.append(p.text.strip())
        for item in p.getchildren():
            ret.extend(self.plain(item))
        if p.tail:
            ret.append(p.tail.strip())
        return ret

    def programlisting(self, p: Element) -> [Md]:
        ret = []
        # programlisting
        if p.tag == 'programlisting':
            got_lang = False
            code = MdCodeBlock([])
            for codeline in p.findall('codeline'):
                line = ''
                for highlight in codeline.findall('highlight'):
                    if not got_lang and len(highlight.getchildren()) == 0 and highlight.text is not None and highlight.text.startswith('{') and highlight.text.endswith('}'):
                        lang = highlight.text[1:-1]
                        code.set_lang(lang)
                        got_lang = True
                        continue
                    else:
                        if highlight.text is not None:
                            line += highlight.text
                        for c in highlight.getchildren():
                            if c.tag == 'sp':
                                line += ' '
                            if c.text:
                                line += c.text
                            if c.tail:
                                line += c.tail
                code.append(line)
            ret.append(Text('\n'))
            ret.append(code)
        return ret

    def paras(self, p: Element, italic: bool = False) -> [Md]:
        ret = []
        if p is None:
            return ret
        if p.text:
            if italic:
                ret.append(MdItalic([Text(p.text.strip())]))
                ret.append(Text(' '))
            else:
                ret.append(Text(p.text))
        for item in p.getchildren():
            # para
            if item.tag == 'para':
                ret.append(MdParagraph(self.paras(item)))
                ret.append(Text('\n'))

            # image
            elif item.tag == 'image':
                url = item.get('name')
                ret.append(MdImage(url))

            # computeroutput
            elif item.tag == 'computeroutput':
                text = []
                if item.text:
                    text.append(item.text)
                for i in item.getchildren():
                    text.extend(self.plain(i))
                ret.append(Code(' '.join(text)))

            # programlisting
            elif item.tag == 'programlisting':
                ret.extend(self.programlisting(item))

            # table
            elif item.tag == 'table':
                t = MdTable()
                for row in item.findall('row'):
                    r = MdTableRow([])
                    for cell in row.findall('entry'):
                        for para in cell.findall('para'):
                            r.append(MdTableCell(self.paras(para)))
                    t.append(r)
                ret.append(t)

            # blockquote
            elif item.tag == 'blockquote':
                b = MdBlockQuote([])
                for para in item.findall('para'):
                    b.extend(self.paras(para))
                ret.append(b)

            # heading
            elif item.tag == 'heading':
                ret.append(MdHeader(int(item.get('level')), self.paras(item)))

            # orderedlist
            elif item.tag == 'orderedlist' or item.tag == 'itemizedlist':
                lst = MdList([])
                for listitem in item.findall('listitem'):
                    i = MdParagraph([])
                    for para in listitem.findall('para'):
                        i.extend(self.paras(para))
                    lst.append(i)
                ret.append(lst)

            # Reference
            elif item.tag == 'ref':
                refid = item.get('refid')
                try:
                    ref = self.cache.get(refid)
                    if italic:
                        if item.text:
                            ret.append(MdLink([MdItalic([MdBold([Text(item.text)])])], ref.url))
                        else:
                            ret.append(MdLink([MdItalic([MdBold([Text(ref.get_full_name())])])], ref.url))
                    else:
                        if item.text:
                            ret.append(MdLink([MdBold([Text(item.text)])], ref.url))
                        else:
                            ret.append(MdLink([MdBold([Text(ref.get_full_name())])], ref.url))
                except:
                    if item.text:
                        ret.append(Text(item.text))

            # sect1:
            elif item.tag == 'sect1':
                title = item.find('title').text
                ret.append(MdHeader(2, [Text(title)]))
                ret.extend(self.paras(item))

            # sect2:
            elif item.tag == 'sect2':
                title = item.find('title').text
                ret.append(MdHeader(3, [Text(title)]))
                ret.extend(self.paras(item))

            # sect3:
            elif item.tag == 'sect3':
                title = item.find('title').text
                ret.append(MdHeader(4, [Text(title)]))
                ret.extend(self.paras(item))

            # sect4:
            elif item.tag == 'sect4':
                title = item.find('title').text
                ret.append(MdHeader(5, [Text(title)]))
                ret.extend(self.paras(item))

            # sect5:
            elif item.tag == 'sect5':
                title = item.find('title').text
                ret.append(MdHeader(6, [Text(title)]))
                ret.extend(self.paras(item))

            # variablelist
            elif item.tag == 'variablelist':
                varlistentry = item.find('varlistentry')
                
                ret.append(MdHeader(4, self.paras(varlistentry.find('term'))))

                term = varlistentry.find('term')
                for listitem in item.findall('listitem'):
                    for para in listitem.findall('para'):
                        ret.append(MdParagraph(self.paras(para)))

            # parameterlist
            elif item.tag == 'parameterlist':
                parameteritems = item.findall('parameteritem')
                lst = MdList([])
                for parameteritem in parameteritems:
                    name = parameteritem.find('parameternamelist').find('parametername')
                    description = parameteritem.find('parameterdescription').findall('para')
                    par = MdParagraph([])
                    par.append(MdBold(self.paras(name)))
                    par.append(Text(' '))
                    for ip in description:
                        par.extend(self.paras(ip))
                    lst.append(par)
                ret.append(Br())
                ret.append(MdBold([Text(SIMPLE_SECTIONS[item.get('kind')])]))
                ret.append(Br())
                ret.append(lst)

            # simplesect
            elif item.tag == 'simplesect':
                kind = item.get('kind')
                if self.hints and self.target == 'vuepress' and kind in SIMPLE_SECTIONS_HINTS_VUEPRESS:
                    ret.append(Br())
                    children = []
                    for sp in item.findall('para'):
                        children.extend(self.paras(sp))
                        children.append(Br())
                    ret.append(MdHint(children, SIMPLE_SECTIONS_HINTS_VUEPRESS[kind], SIMPLE_SECTIONS[kind]))

                else:
                    ret.append(Br())
                    ret.append(MdBold([Text(SIMPLE_SECTIONS[kind])]))
                    if kind != 'see':
                        ret.append(Br())
                    else:
                        ret.append(Text(' '))

                    for sp, has_more in lookahead(item.findall('para')):
                        ret.extend(self.paras(sp))
                        if kind == 'see':
                            if has_more:
                                ret.append(Text(', '))
                        else:
                            ret.append(Br())

            # xrefsect
            elif item.tag == 'xrefsect':
                xreftitle = item.find('xreftitle')
                xrefdescription = item.find('xrefdescription')
                kind = xreftitle.text.lower()
                if self.hints and self.target == 'vuepress' and kind in SIMPLE_SECTIONS_HINTS_VUEPRESS:
                    children = []
                    for sp in xrefdescription.findall('para'):
                        children.extend(self.paras(sp))
                        children.append(Br())
                    ret.append(MdHint(children, SIMPLE_SECTIONS_HINTS_VUEPRESS[kind], SIMPLE_SECTIONS[kind]))
                else:
                    ret.append(Br())
                    ret.append(MdBold(self.paras(xreftitle)))
                    ret.append(Br())
                    for sp in xrefdescription.findall('para'):
                        ret.extend(self.paras(sp))
                        ret.append(Br())

            # Hard link
            elif item.tag == 'ulink':
                ret.append(MdLink(self.paras(item), item.get('url')))

            # Bold
            elif item.tag == 'bold':
                ret.append(MdBold(self.paras(item)))

            # Emphasis
            elif item.tag == 'emphasis':
                ret.append(MdItalic(self.paras(item)))

            # End of the item text
            if item.tail:
                if italic:
                    ret.append(Text(' '))
                    ret.append(MdItalic([Text(item.tail.strip())]))
                else:
                    ret.append(Text(item.tail))
        return ret



# This template is used inside of member.py
TEMPLATE = """
{% if target == 'gitbook' %}
### <a href="#{{node.anchor}}" id="{{node.anchor}}">{{node.kind.value}} {{node.name_long if node.is_group else node.name_short}} {{node.overload_suffix}}</a>
{% else %}
### {{node.kind.value}} {{node.name_long if node.is_group else node.name_short}} {{node.overload_suffix}}
{% endif %}

```cpp
{{node.codeblock}}
```

{% if node.has_details -%}
{{node.details}}
{%- endif -%}

{% if node.reimplements %}
Implements [*{{node.reimplements.name_long}}*]({{node.reimplements.url}})
{% endif %}
"""



TEMPLATE = """
# Enum List

Here are the enums with brief descriptions:


{% for node in nodes recursive %}
{%- if node.is_enum and (node.brief or node.details) %}
* **{{node.kind.value}}** [**{{node.name_short}}**]({{node.refid+'.md'}}) {{node.brief}}
{%- endif %}
{%- if node.has_children %}
  {{- loop(node.children) }}
{%- endif -%}
{%- endfor %}
"""




class Cache:
    def __init__(self):
        self.cache = {}
        self.cache_on_name = {}

    def add(self, key: str, value):
        self.cache[key] = value

    def get(self, key: str):
        if key in self.cache:
            return self.cache[key]
        else:
            raise IndexError('Key: ' + key + ' not found in cache!')

    def add_onname(self, key: str, value):
        self.cache_on_name[key] = value #add by jiangzq

    def get_on_name(self, key: str):
        if key in self.cache_on_name:
            return self.cache_on_name[key]
        else:
            #print('aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa')
            print(key)
            #for key,value in self.cache_on_name.items():
            #    print(key,":",value.name)
            #raise IndexError('Key: ' + key + ' not found in cache!')
            print('Worning: Key: ' + key + ' not found in cache!')



TEMPLATE = """
# Class Hierarchy

This inheritance list is sorted roughly, but not completely, alphabetically:

{% for node in classes recursive %}
{% if node.is_resolved -%}
* **{{node.kind.value}}** [**{{node.name_long}}**]({{node.url}}) {{node.brief}}
{%- else -%}
* **{{node.kind.value}}** **{{node.name_long}}**
{%- endif %}
{%- if node.derived_classes %}
  {{- loop(node.derived_classes)|indent(2, true) }}
{%- endif %}
{%- endfor %}
"""



# This template is used inside of member.py
TEMPLATE = """
{%- if node.has(visibility, kinds, static) -%}
{%- if parent is none -%}
## {{title}}
{%- else -%}
## {{title}} inherited from [{{node.name_long}}]({{node.url}})
{%- endif %}

| Type | Name |
| ---: | :--- |
{% for member in node.query(visibility, kinds, static) -%}
{%- if member.is_group -%}
| module | [**{{member.title}}**]({{member.url}}) {{member.suffix}}<br>{{member.brief}} |
{%- elif member.is_file -%}
| file | [**{{member.name_short}}**]({{member.url}}) {{member.suffix}}<br>{{member.brief}} |
{%- elif member.is_dir -%}
| dir | [**{{member.name_short}}**]({{member.url}}) {{member.suffix}}<br>{{member.brief}} |
{%- elif member.is_namespace -%}
| {{member.kind.value}} | [**{{member.name_long if node.is_group else member.name_short}}**]({{member.url}}) {{member.suffix}}<br>{{member.brief}} |
{%- elif member.is_class or member.is_interface or member.is_struct -%}
| {{member.kind.value}} | [**{{member.name_long if node.is_group else member.name_short}}**]({{member.url}}) {{member.suffix}}<br>{{member.brief}} |
{%- elif member.is_enum or member.is_function or member.is_variable or member.is_union or member.is_typedef -%}
| {{member.prefix}} {{member.type}} | [**{{member.name_long if node.is_group else member.name_short}}**]({{member.url}}) {{member.params}} {{member.suffix}}<br>{{member.brief}} |
{%- else -%}
| {{member.prefix}} {{member.type}} | [**{{member.name_long if node.is_group else member.name_short}}**]({{member.url}}) {{member.params}} {{member.suffix}}<br>{{member.brief}} |
{%- endif %}
{% endfor -%}

{%- endif -%}
"""



import os
from xml.etree import ElementTree
from doxybook.node import Node
from doxybook.constants import Kind, Visibility
from doxybook.cache import Cache
from doxybook.xml_parser import XmlParser


class Doxygen:
    def __init__(self, input_path: str, parser: XmlParser, cache: Cache, \
            doc_type = 'all', rel_path = ''):
        path = os.path.join(input_path, 'index.xml')
        self.parser = parser
        self.cache = cache
        self.rootnodes = []
        self.groupsnodes =[] 
        self.filesnodes =[] 
        self.pagesnodes =[] 
        self.index_paths = []
        self.doc_type = doc_type
        self.rel_path = rel_path
        #if os.path.exists(path):
        #    self.index_paths.append(path)
        #else:
        #    dirs = os.listdir(input_path)
        #    for dir_s in dirs:
        #        dir_s = os.path.join(input_path,dir_s)
        #        if os.path.isdir(dir_s):
        #            self.index_paths.append(os.path.join(dir_s, 'index.xml'))
        self.get_index(input_path)
        for path in self.index_paths:
            print('Loading XML from: ' + path)
            xml = ElementTree.parse(path).getroot()
            index_path = os.path.dirname(path)

            self.root = Node('root', None, self.cache, self.parser, None, \
                    doc_type = self.doc_type, rel_path = self.rel_path)
            self.groups = Node('root', None, self.cache, self.parser, None, \
                    doc_type = self.doc_type, rel_path = self.rel_path)
            self.files = Node('root', None, self.cache, self.parser, None, \
                    doc_type = self.doc_type, rel_path = self.rel_path)
            self.pages = Node('root', None, self.cache, self.parser, None, \
                    doc_type = self.doc_type, rel_path = self.rel_path)
            

            for compound in xml.findall('compound'):
                kind = Kind.from_str(compound.get('kind'))
                refid = compound.get('refid')
                if kind.is_language():
                    node = Node(os.path.join(index_path, refid + '.xml'), None, self.cache, self.parser, self.root, \
                            doc_type = self.doc_type, rel_path = self.rel_path)
                    node._visibility = Visibility.PUBLIC
                    self.root.add_child(node)
                if kind == Kind.GROUP:
                    node = Node(os.path.join(index_path, refid + '.xml'), None, self.cache, self.parser, self.root, \
                            doc_type = self.doc_type, rel_path = self.rel_path)
                    node._visibility = Visibility.PUBLIC
                    self.groups.add_child(node)
                if kind == Kind.FILE :
                #if kind == Kind.FILE or kind == Kind.DIR:
                    node = Node(os.path.join(index_path, refid + '.xml'), None, self.cache, self.parser, self.root, \
                            doc_type = self.doc_type, rel_path = self.rel_path)
                    node._visibility = Visibility.PUBLIC
                    self.files.add_child(node)
                if kind == Kind.PAGE:
                    node = Node(os.path.join(index_path, refid + '.xml'), None, self.cache, self.parser, self.root, \
                            doc_type = self.doc_type, rel_path = self.rel_path)
                    node._visibility = Visibility.PUBLIC
                    self.pages.add_child(node)

            print('Deduplicating data... (may take a minute!)')
            for i, child in enumerate(self.root.children.copy()):
                self._fix_duplicates(child, self.root, [])

            for i, child in enumerate(self.groups.children.copy()):
                self._fix_duplicates(child, self.groups, [Kind.GROUP])

            for i, child in enumerate(self.files.children.copy()):
                self._fix_duplicates(child, self.files, [Kind.FILE, Kind.DIR])

            self._fix_parents(self.files)

            print('Sorting...')
            #self._recursive_sort(self.root)
            #self._recursive_sort(self.groups)
            #self._recursive_sort(self.files)
            #self._recursive_sort(self.pages)
            self.rootnodes.append(self.root)
            self.groupsnodes.append(self.groups)
            self.filesnodes.append(self.files)
            self.pagesnodes.append(self.pages)
            #if path == "/home/gx/bbb/ccc/./XML/demuxhal/index.xml":
            #    break
            #if path == '/home/gx/bbb/ccc/./XML/deshal/index.xml':
            #    break
            #if path == '/home/gx/bbb/ccc/./XML/fehal/index.xml':
            #    break

    #def get_index(self, path):
    #    if os.path.exists(os.path.join(path,'index.xml')):
    #        self.index_paths.append(path)
    #        return
    #    else:
    #        dirs = os.listdir(path)
    #        for dir_s in dirs:
    #            dir_s = os.path.join(path,dir_s)
    #            if os.path.isdir(dir_s):
    #                if os.path.exists(os.path.join(dir_s, 'index.xml')):
    #                    self.index_paths.append(os.path.join(dir_s, 'index.xml'))
    #                else:
    #                    self.get_index(dir_s)
    def list_node(self, node, leval = 0):
        n = leval
        print('\t'*n, node.kind, node.name, 'now')
        if node.has_children:
            n += 1
            for child in node.children:
                print('\t'*n, child.kind, child.name, 'children')
                if child.has_children:
                    self.list_node(child, n)


    def get_index(self, path):
        if os.path.exists(os.path.join(path,'index.xml')):
            index_full_path = os.path.join(path, 'index.xml')
            if index_full_path not in self.index_paths:
                self.index_paths.append(index_full_path)
        dirs = os.listdir(path)
        for dir_s in dirs:
            dir_s = os.path.join(path,dir_s)
            if os.path.isdir(dir_s):
                if os.path.exists(os.path.join(dir_s, 'index.xml')):
                    index_full_path = os.path.join(dir_s, 'index.xml')
                    if index_full_path not in self.index_paths:
                        self.index_paths.append(index_full_path)
                self.get_index(dir_s)


    def _fix_parents(self, node: Node):
        if node.is_dir or node.is_root:
            for child in node.children:
                if child.is_file:
                    child._parent = node
                if child.is_dir:
                    self._fix_parents(child)

    def _recursive_sort(self, node: Node):
        node.sort_children()
        for child in node.children:
            self._recursive_sort(child)

    def _is_in_root(self, node: Node, root: Node):
        for child in root.children:
            if node.refid == child.refid:
                return True
        return False

    def _remove_from_root(self, refid: str, root: Node):
        for i, child in enumerate(root.children):
            if child.refid == refid:
                root.children.pop(i)
                return

    def _fix_duplicates(self, node: Node, root: Node, filter: [Kind]):
        for child in node.children:
            if len(filter) > 0 and child.kind not in filter:
                continue
            if self._is_in_root(child, root):
                self._remove_from_root(child.refid, root)
            self._fix_duplicates(child, root, filter)

    def print(self):
        #for node in self.root.children:
        #    self.print_node(node, '')
        #for node in self.groups.children:
        #    self.print_node(node, '')
        #for node in self.files.children:
        #    self.print_node(node, '')
        pass

    def print_node(self, node: Node, indent: str):
        print(indent, node.kind, node.name, node.url)
        #if node.kind == Kind.FUNCTION and node.name == 'libusb_interrupt_transfer':
        #    print(node.brief)
        for child in node.children:
            self.print_node(child, indent + '  ')



#!/usr/bin/env python3

import sys
import os
import re
import shutil
import argparse

class GenXML():
    def __init__(self, path, outputpath, is_server=None, path_type = 'all'):
        self.path = path #doc_path
        self.outputpath = outputpath
        if not os.path.exists(self.outputpath):
            os.system('mkdir -p '  + self.outputpath)
        #if not os.path.exists(os.path.join(self.outputpath, 'XML')):
        #    #@shutil.rmtree(os.path.join(self.outputpath, 'XML'))
        #    os.mkdir(os.path.join(self.outputpath, 'XML'))
        if not os.path.exists(os.path.join(self.outputpath, 'XML')):
            os.mkdir(os.path.join(self.outputpath, 'XML'))
        self.available_paths = {}
        self.is_server = is_server
        self.path_type = path_type

    def getdocdirs(self):
        absdocpath = os.path.join(self.path, 'doc')
        if os.path.exists(absdocpath):
            self.available_paths[os.path.basename(self.path)] = absdocpath
        else:
            modules = os.listdir(self.path)
            for module in modules:
                absdocpath = os.path.join(self.path, module, 'doc')
                if os.path.exists(absdocpath):
                    self.available_paths[module] = absdocpath

    def getmoduledoxygenconf(self,path):
        cfgdirs = []
        for dirpath,dirnames,filenames in os.walk(path):
            if self.is_server == 'yes':
                cmd = 'cd ' + dirpath + ' && git checkout . && cd -'
                os.system(cmd)
            doxygenconf = os.path.join(dirpath, 'doxygen.cfg')
            if os.path.exists(doxygenconf):
                cfgdirs.append(dirpath)
        return cfgdirs
    
    def have_no_dir_in_xml(self, targetdir, cfgdir):
        if not os.path.exists(targetdir):
            os.makedirs(targetdir)
        targetdir = re.sub('/', '\/', targetdir)

        cmd = "sed -i \'s/XML_OUTPUT \+=.\+$/XML_OUTPUT = " +\
                targetdir + '/g\' ' + cfgdir + '/doxygen.cfg'
        os.system(cmd)
        os.system('doxygen doxygen.cfg')
        no_dir = 1
        xmlpath = os.path.join(self.outputpath, 'XML')
        for i in os.listdir(xmlpath):
            path = os.path.join(xmlpath, i)
            if os.path.isdir(path):
                no_dir = 0

        os.system('rm -rf ' + xmlpath)
        return no_dir

    def getmodulexml(self, path):
        cfgdirs = self.getmoduledoxygenconf(path)
        if not cfgdirs:
            return None
        xmlpath = os.path.join(self.outputpath, 'XML')
        for cfgdir in cfgdirs:
            os.chdir(cfgdir)
            targetdir = re.sub(self.path, xmlpath, cfgdir)
            #if self.path_type == "all":
            #if os.path.basename(targetdir) == 'doc':
            #    #targetdir = re.sub('doc', '', targetdir)
            #else:
            #    targetdir = re.sub('doc/', '', targetdir)
            tmp = targetdir.split('/')
            new_tmp = []
            for i in tmp:
                if not i == 'doc':
                    new_tmp.append(i)
            targetdir = '/'.join(new_tmp)
            if not os.path.exists(targetdir):
                os.makedirs(targetdir)
            targetdir = re.sub('/', '\/', targetdir)

            cmd = "sed -i \'s/XML_OUTPUT \+=.\+$/XML_OUTPUT = " +\
                    targetdir + '/g\' ' + cfgdir + '/doxygen.cfg'
            os.system(cmd)
            os.system('doxygen doxygen.cfg')
        return cfgdirs

    def GetXml2(self):
        modules = self.getdocdirs()
        config_dirs = []
        for module, docpath in self.available_paths.items():
            cfgdirs = self.getmodulexml(docpath)
            if not cfgdirs:
                continue
            config_dirs.extend(cfgdirs)
        return self.available_paths, config_dirs

    def GetXml(self):
        self.getdocdirs()
        config_dirs = []
        if self.path_type == 'all':
            for module, docpath in self.available_paths.items():
                print(module, docpath)
                cfgdirs = self.getmodulexml(docpath)
                print(cfgdirs)
                if not cfgdirs:
                    continue
                config_dirs.extend(cfgdirs)
        elif self.path_type == 'part':
            cfgdirs = self.getmodulexml(self.path)
            config_dirs.extend(cfgdirs)
        return config_dirs

#doxybook -i xml -o book/solutiontest -s book/SUMMARY.md -t gitbook
class GenMD():
    def __init__(self, start_path, bookpath, source_paths, xmlpath, \
            path_type = 'all', link_path = '', now_dir = '', rel_path = '', enable_include_link = False):
        self.bookpath = bookpath
        self.source_paths = source_paths
        self.xmlpath = xmlpath
        self.start_path = start_path #doc_path
        self.path_type = path_type
        self.link_path = link_path
        self.exec_dir = now_dir
        self.rel_path = rel_path
        self.enable_include_link = enable_include_link
        if self.link_path.startswith('/opt/'):
            self.tmp_link_path = '/' + '/'.join(self.link_path.strip('/').split('/')[1:])

    def _check_path(self):
        if not os.path.exists(self.bookpath):
            os.makedirs(self.bookpath)

    def is_need_copy(self, path):
        for root,dirs,files in os.walk(path):
            for file_name in files:
                if 'doxygen.cfg' in file_name:
                    return False
        return True

    def copymdfile(self, path):
        #targetdir = re.sub(self.start_path,'',path)
        targetdir = os.path.relpath(path, self.start_path)
        #if self.path_type == "all":
        #if os.path.basename(path) == 'doc':
        #    targetdir = re.sub('doc','',targetdir)
        #else:
        #    targetdir = re.sub('doc/','',targetdir)
        tmp = targetdir.split('/')
        new_tmp = []
        for i in tmp:
            if not i == 'doc':
                new_tmp.append(i)
        targetdir = '/'.join(new_tmp)
        targetdir = targetdir.strip("/")
        mdfiles = os.listdir(path)
        if not os.path.exists(os.path.join(self.bookpath, targetdir)):
            os.makedirs(os.path.join(self.bookpath, targetdir))
        for mdfile in mdfiles:
            if mdfile[0] == '.' or mdfile == 'doxygen.cfg':
                continue
            if mdfile == 'docs' and self.path_type == 'part':
                continue
            if os.path.isfile(os.path.join(path,mdfile)):
                if mdfile == 'example.md' and self.path_type == 'all':
                    examplemd = os.path.join(path,mdfile)
                    fd = open(examplemd, 'r')
                    data = fd.read()
                    fd.close()
                    data = re.sub('^[ \t]*\n','',data)
                    if len(data) >= 3:
                        info = data.split('\n')
                        if len(info) >= 2:
                            project = info[0].lower()
                            module = info[1]
                            projects = ['gxbus','api']
                            if project in projects:
                                projectpath = os.path.join(self.start_path,\
                                        'gxtest',project)
                                cdcmd = ' cd ' + projectpath
                                buildcmd = './build config -k example -m ' + module
                                cpcmd = 'cp -f example.md ' + examplemd
                                cmd = cdcmd + ';'+ buildcmd + ';' + cpcmd
                                os.system(cmd)
                elif mdfile.endswith('.md') or mdfile.endswith('.MD'):
                    cmd = 'cp  \"' + os.path.join(path,mdfile) + '\" ' + \
                            os.path.join(self.bookpath, targetdir)
                    os.system(cmd)
            elif mdfile == 'images':
                cmd = 'cp -rf \"' + os.path.join(path,mdfile) + '\" ' + \
                        os.path.join(self.bookpath, targetdir)
                os.system(cmd)
            #elif self.is_need_copy(mdfile):
            #    cmd = 'cp -rf \"' + os.path.join(path,mdfile) + '\" ' + \
            #            os.path.join(self.bookpath, targetdir)
            #    os.system(cmd)
            else:
                self.copymdfile(os.path.join(path,mdfile))

    def _touch_summary(self, md_dict):
        #fd.write('* [' + name + '](index.md)\n')
        summary =  os.path.join(self.bookpath, 'SUMMARY.md')
        with open(summary, 'w') as fd:
            fd.write('# Summary \n\n')
            for i in md_dict.values():
                if 'index' in i:
                    md_name = os.path.relpath(os.path.join(i['path'], 'index.md'), self.start_path)
                    fd.write('* [' + md_name + '](' + md_name + ')\n')
                    if 'files' in i:
                        for f in i['files']:
                            md_name = os.path.relpath(os.path.join(i['path'], f), self.start_path)
                            fd.write('  * [' + md_name + '](' + md_name + ')\n')
                else:
                    if 'files' in i:
                        for f in i['files']:
                            md_name = os.path.relpath(os.path.join(i['path'], f), self.start_path)
                            fd.write('* [' + md_name + '](' + md_name + ')\n')

    def _touch_readme(self):
        readme =  os.path.join(self.bookpath, 'readme.md')
        with open(readme, 'w') as fd:
            fd.write('  ')

    def _find_all_md_path(self, in_dir):
        path_list = []
        for i in os.listdir(in_dir):
            path = os.path.join(in_dir, i)
            if os.path.isdir(path):
                pass
            else:
                if path.endswith(".md") and not path.endswith('/index.md'):
                    path_list.append(path)
        return path_list

    def before_work(self):
        self._check_path()
        md_dict = {}
        if self.path_type == 'all':
            pass
        else:
            for cfgdir in self.source_paths:
                path_msg = cfgdir.split('/doc')[1].strip('/')
                md_dict[path_msg] = {
                        'path':cfgdir,
                        }
                if os.path.exists(os.path.join(cfgdir, 'index.md')):
                    md_dict[path_msg]['index'] = 'index.md'
                other_md_files = self._find_all_md_path(cfgdir)
                if len(other_md_files) > 0:
                    md_dict[path_msg]['files'] = other_md_files
            self._touch_summary(md_dict)
            self._touch_readme()

    def GenMD(self, is_all):
        self.before_work()
        if self.path_type == 'all':
            cmd = "doxybook -i " + self.xmlpath + " -o " + self.bookpath + ' -s ' + \
                self.bookpath + '/SUMMARY.md -t gitbook -a ' + is_all
        elif self.path_type == 'part':
            if self.enable_include_link:
                if self.link_path.startswith('/opt/'):
                    docref_path = os.path.join(self.tmp_link_path, self.rel_path)
                else:
                    docref_path = os.path.join(self.link_path, self.rel_path)
            else:
                docref_path = self.rel_path
            cmd = "doxybook -i " + self.xmlpath + " -o " + self.bookpath + ' -s ' + \
                self.bookpath + '/SUMMARY.md -t gitbook -a ' + is_all + ' -p part -l ' + docref_path
        os.system(cmd)
        #for cfgdir in self.source_paths:
        #    self.copymdfile(cfgdir)
        self.copymdfile(self.start_path)

    def check_file(self, file_name):
        #if os.path.exists(os.path.join(self.exec_dir, self.link_path, \
        #        "_gen_output", file_name)):
        if not self.link_path.startswith('/'):
            if os.path.exists(os.path.join(self.exec_dir, self.link_path, file_name)):
                add_cmd = "cat " + os.path.join(self.bookpath, file_name) + " >> " \
                        + os.path.join(self.exec_dir, self.link_path, file_name)
                #        + os.path.join(self.exec_dir, self.link_path, "_gen_output", file_name)
                os.system(add_cmd)
            else:
                cp_cmd = "cp " + os.path.join(self.bookpath, file_name) + " " \
                        + os.path.join(self.exec_dir, self.link_path, file_name)
                        #+ os.path.join(self.exec_dir, self.link_path, "_gen_output", file_name)
                os.system(cp_cmd)
        else:
            if os.path.exists(os.path.join(self.link_path, file_name)):
                add_cmd = "cat " + os.path.join(self.bookpath, file_name) + " >> " \
                        + os.path.join(self.link_path, file_name)
                os.system(add_cmd)
            else:
                cp_cmd = "cp " + os.path.join(self.bookpath, file_name) + " " \
                        + os.path.join(self.link_path, file_name)
                os.system(cp_cmd)
    
    def modify_summary_path(self):
        summary =  os.path.join(self.bookpath, 'SUMMARY.md')
        new_summary =  os.path.join(self.bookpath, 'SUMMARY2.md')
        include_summary =  os.path.join(self.bookpath, 'include_summary.md')
        fd = open(summary, 'r')
        fn = open(new_summary, 'w')
        fi = open(include_summary, 'w')
        while 1:
            line = fd.readline()
            if not line:
                break
            if line.find('](') >= 0:
                line = line.strip('\n')
                index = line.split('[')[0]
                name = line.split('[')[1].split('](')[0]
                link = line.split('](')[1][:-1]
                if link.startswith('doc/'):
                    link = link[4:]
                n_line = index +  '[' + name + '](' + os.path.join(\
                        self.rel_path, 'book', link.strip('/')) + ')\n'
                #i_line = index +  '[' + name + '](' + os.path.join(self.link_path, '_gen_output', \
                if self.link_path.startswith('/opt/'):
                    i_line = index +  '[' + name + '](' + os.path.join(self.tmp_link_path, \
                        self.rel_path, 'book', link.strip('/')) + ')\n'
                else:
                    i_line = index +  '[' + name + '](' + os.path.join(self.link_path, \
                        self.rel_path, 'book', link.strip('/')) + ')\n'
                fn.write(n_line)
                fi.write(i_line)
            else :
                fn.write(line)
                fi.write(line)
        fd.close()
        fn.close()
        fi.close()
        mv_cmd = 'mv ' + new_summary + ' ' + summary
        os.system(mv_cmd)


def parse_options():
    parser = argparse.ArgumentParser(description='Convert doxygen XML output \
            into GitBook or Vuepress markdown output.')
    parser.add_argument('-d', '--dir',
        help='Path of module',
    )
    parser.add_argument('-t', '--type',
        help='Path Type',
    )
    parser.add_argument('-r', '--root_dir',
        help='root_dir of book, such as xxx/goxceed/v1.9-dev',
    )
    parser.add_argument('-l', '--link',
        help='doc output path',
    )
    parser.add_argument('-i', '--input', 
        type=str, 
        help='Path to doxygen goxceed folder',
    )
    parser.add_argument('-s', '--server', 
        type=str, 
        help='Run on server',
    )
    parser.add_argument('-a', '--allgenerator', 
        type=str, 
        help='Is generator all *.md in Summary',
    )
    parser.add_argument('-g', '--include', action='store_const', const = True,
        default = False,
        help='enable in include_summary'
    )


    args = parser.parse_args()

    return args


def main():
    if len(sys.argv) == 1:
        print("")
        print("Please input the goxceed path!")
        print("")
        print(" genbook ../goxceed")
        print("")
        sys.exit(1)
    args = parse_options()
    if args.dir != None:
        now_dir = os.getcwd()
        doc_path = args.dir
        if not doc_path.startswith('/'):
            doc_path = os.path.join(now_dir, args.dir)

        link_path = args.link
        #if not link_path.startswith('/'):
        #    link_path = os.path.join(now_dir, args.link)

        root_path = args.root_dir
        if root_path == None:
            root_path = doc_path
        elif not root_path.startswith('/'):
            root_path = os.path.join(now_dir, args.root_dir)

        if root_path != doc_path:
            if os.path.basename(root_path) == 'doc':
                root_path = re.sub('doc', '', root_path)
            else:
                root_path = re.sub('doc/', '', root_path)


        #path_type = 'all'
        #if doc_path != root_path:
        #    path_type = 'part'
        enable_include_link = args.include
        path_type = args.type

        rel_path = os.path.relpath(doc_path, root_path)
        if os.path.basename(rel_path) == 'doc':
            rel_path = re.sub('doc', '', rel_path)
        else:
            rel_path = re.sub('doc/', '', rel_path)

        doc_path = os.path.abspath(doc_path)
        #outpath = os.path.join(now_dir, link_path, '_gen_output', rel_path)
        if not link_path.startswith('/'):
            outpath = os.path.join(now_dir, link_path, rel_path)
        else:
            outpath = os.path.join(link_path, rel_path)
        os.chdir(doc_path)
        #print('link_path : ' +  link_path)
        #print('root_path : ' +  root_path)
        #print('doc_path : ' +  doc_path)
        #print('rel_path : ' + rel_path)
        #print('outpath : ' + outpath)
        genxml = GenXML(doc_path, outpath, path_type = path_type)
        cfgdirs = genxml.GetXml()
        bookoutpath = os.path.join(outpath, 'book')
        print(bookoutpath)
        genmd = GenMD(doc_path, bookoutpath, cfgdirs, outpath+'/XML', \
                path_type, link_path, now_dir, rel_path, enable_include_link)
        genmd.GenMD('yes')
        genmd.modify_summary_path()
        genmd.check_file('SUMMARY.md')
        genmd.check_file('include_summary.md')
        genmd.check_file('readme.md')
    elif args.input != None:
        doc_path = args.input
        is_server = args.server
        allgenerator = args.allgenerator
        doc_path = os.path.abspath(doc_path)
        outpath = doc_path
        outpath = os.path.join(outpath, 'docs')
        os.chdir(doc_path)
        genxml = GenXML(doc_path, outpath, is_server)
        modules, cfgdirs = genxml.GetXml2()
        if is_server == 'yes':
            for cfgdir in cfgdirs:
                os.remove(os.path.join(cfgdir,'doxygen.cfg'))
        for module, modpath in modules.items():
            if not modpath in cfgdirs:
                cfgdirs.append(modpath)
        #if is_server == 'yes':
        #    for cfgdir in cfgdirs:
        #        os.remove(os.path.join(cfgdir,'doxygen.cfg'))
        bookoutpath = os.path.join(outpath, 'book')
        genmd = GenMD(doc_path, bookoutpath, cfgdirs, outpath+'/XML' )
        genmd.GenMD(allgenerator)


if __name__ == "__main__":
    main()


#for dirpath,dirnames,filenames in os.walk(doc_path+'/gxtest/doc'):
#    print(dirpath, dirnames, filenames)
#print(os.listdir(doc_path))
#
#print(os.getcwd())
##os.chdir()
#print(doc_path)





import argparse
import sys
import os
from doxybook.runner import run


def parse_options():
    parser = argparse.ArgumentParser(description='Convert doxygen XML output into GitBook or Vuepress markdown output.')
    parser.add_argument('-t', '--target', 
        type=str, 
        help='Select the target: Gitbook (default) or Vuepress, for example: "-t vuepress" or "-t gitbook"',
        required=False,
        default='gitbook'
    )
    parser.add_argument('-i', '--input', 
        type=str, 
        help='Path to doxygen generated xml folder',
        required=True
    )
    parser.add_argument('-o', '--output', 
        type=str, 
        help='Path to the destination folder',
        required=True
    )
    parser.add_argument('-s', '--summary', 
        type=str, 
        help='Path to the summary file which contains a link to index.md in the folder pointed by --input (default: false)',
        required=False
    )
    parser.add_argument('-d', '--debug', 
        type=bool, 
        help='Debug the class hierarchy (default: false)',
        required=False,
        default=False
    )
    parser.add_argument('--hints', 
        type=bool, 
        help='(Vuepress only) If set to true, hints will be generated for the sections note, bug, and a warning (default: true)',
        required=False,
        default=True
    )
    parser.add_argument('--ignoreerrors', 
        type=bool, 
        help='If set to true, will continue to generate Markdown files even if an error has been detected (default: false)',
        required=False,
        default=False
    )
    parser.add_argument('-n', '--noexp', action='store_const',const=True,
            default=False,
            help='not create example')

    args = parser.parse_args()

    if args.target != 'gitbook' and args.target != 'vuepress':
        raise Exception('Unknown target: ' + str(args.target))

    if args.target == 'gitbook' and args.summary and not os.path.exists(args.summary):
        raise Exception('The provided summary file does not exist!')

    if os.path.isfile(args.output):
        raise Exception('The target output directory is a file!')

    return args

def main():
    args = parse_options()
    os.makedirs(args.output, exist_ok=True)

    run(
        input=args.input,
        output=args.output,
        target=args.target,
        hints=args.hints,
        debug=args.debug,
        ignore_errors=args.ignoreerrors,
        summary=args.summary,
        no_example = args.noexp
    )



"""
Documentation
-------------
source code from : https://github.com/matusnovak/doxybook

"""

from setuptools import setup, find_packages

long_description = __doc__

def main():
    setup(
        name="GxDoxybook",
        description="Convert C annotation to gitbook",
        keywords="c annotation gitbook",
        long_description=long_description,
        version="1.0.4",
        author="zhaobk",
        author_email="zhaobk@nationalchip.com",
        packages=['doxybook', 'doxybook.templates'],
        install_requires=['Jinja2'],
        package_data={},
        entry_points={
            'console_scripts':[
                'doxybook=doxybook:main',
                'genbook=doxybook.genbook:main',
                ]
            }
    )


if __name__ == "__main__":
    main()



TEMPLATE = """
# {{node.kind.value|title}} {{node.name_long}}
{% if node.has_templateparams %}
**template &lt;{{node.templateparams}}&gt;**
{% endif %}

{% if node.is_group -%}
[**Modules**](modules.md)
{% elif node.is_file or node.is_dir -%}
[**File List**](files.md)
{%- else -%}
[**Class List**](annotated.md)
{%- endif %}
{%- for parent in node.parents -%}
{{'**>**'|indent(1, true)}} [**{{parent.name_long if node.is_group else parent.name_short}}**]({{parent.url}})
{%- endfor %}

```cpp
{{node.codeblock}}
```
{% if node.is_file and node.has_programlisting -%}
[Go to the source code of this file.]({{node.url_source}})
{%- endif %}

{{node.brief}}
{%- if node.has_details -%}
[More...](#detailed-description)
{%- endif %}

{% if node.has_includes -%}
{%- for include in node.includes -%}
* `#include {{include}}`
{% endfor -%}
{%- endif %}

{% if node.has_base_classes %}
Inherits the following classes:
{%- for base in node.base_classes -%}
{%- if base is string %} {{base}}{%- else %} [{{base.name_long}}]({{base.url}}){%- endif -%}
{{ ', ' if not loop.last else '' }}
{%- endfor -%}
{%- endif %}

{% if node.has_derived_classes %}
Inherited by the following classes:
{%- for derived in node.derived_classes -%}
{%- if derived is string %} {{derived}}{%- else %} [{{derived.name_long}}]({{derived.url}}){%- endif -%}
{{ ', ' if not loop.last else '' }}
{%- endfor -%}
{%- endif %}

{{ member_table_template.render({'target': target, 'node': node, 'parent': None, 'title': 'Files', 'visibility': 'public', 'kinds': ['file'], 'static': False}) }}
{{ member_table_template.render({'target': target, 'node': node, 'parent': None, 'title': 'Directories', 'visibility': 'public', 'kinds': ['dir'], 'static': False}) }}
{{ member_table_template.render({'target': target, 'node': node, 'parent': None, 'title': 'Modules', 'visibility': 'public', 'kinds': ['group'], 'static': False}) }}
{{ member_table_template.render({'target': target, 'node': node, 'parent': None, 'title': 'Namespaces', 'visibility': 'public', 'kinds': ['namespace'], 'static': False}) }}
{{ member_table_template.render({'target': target, 'node': node, 'parent': None, 'title': 'Classes', 'visibility': 'public', 'kinds': ['class', 'struct', 'interface'], 'static': False}) }}

{%- for visibility in ['public', 'protected'] -%}
{%- for query in [['types', ['enum', 'union', 'typedef']], ['attributes', ['variable']], ['functions', ['function']]] -%}
{%- for static in [['', False], ['static ', True]] %}
{{ member_table_template.render({'target': target, 'node': node, 'parent': None, 'title': visibility|title + ' ' + static[0]|title + query[0]|title, 'visibility': visibility, 'kinds': query[1], 'static': static[1]}) }}
{%- for child in node.base_classes recursive -%}{%- if child is not string %}
{{ member_table_template.render({'target': target, 'node': child, 'parent': node, 'title': visibility|title + ' ' + static[0]|title + query[0]|title, 'visibility': visibility, 'kinds': query[1], 'static': static[1]}) }}
{{- loop(child.base_classes)}}
{%- endif -%}{%- endfor -%}
{%- endfor -%}
{%- endfor -%}
{%- endfor -%}


"""



import sys
import os
import re
from doxybook.doxygen import Doxygen
from doxybook.generator import Generator
from doxybook.xml_parser import XmlParser
from doxybook.cache import Cache
from doxybook.constants import Kind

def list_node(node, leval = 0):
    n = leval
    print('\t'*n, node.kind, node.name, 'now')
    if node.has_children:
        n += 1
        for child in node.children:
            print('\t'*n, child.kind, child.name, 'children')
            if child.has_children:
                list_node(child, n)

def run(output: str, 
        input: str, 
        target: str = 'gitbook',
        hints: bool = True, 
        debug: bool = True, 
        ignore_errors: bool = False,
        summary: str = None,
        allgenerator: str = None,
        doc_type = 'all',
        rel_path = '',
        ):

    os.makedirs(output, exist_ok=True)

    no_example = False
    if doc_type == 'part':
        no_example = True
    cache = Cache()
    parser = XmlParser(cache=cache, target=target, hints=hints)
    doxygen = Doxygen(input, parser, cache, doc_type = doc_type, rel_path = rel_path)
    #print('root.........................................')
    #for node in doxygen.rootnodes[0].children:
    #    print(node.name, node.kind)
    #print('groups.........................................')
    #for node in doxygen.groupsnodes[0].children:
    #    print(node.name, node.kind)
    #    if node.has_children:
    #        for child in node.children:
    #            print(node.name, node.kind)
    #sys.exit(0)
    
    #print("runing ..................................")
    #doxygen.print()
    if debug:
        doxygen.print()
    for index,v in enumerate(doxygen.index_paths):
        output_dir = os.path.join(output,os.path.basename(os.path.dirname(v)))
        if '/XML/' in v:
            xmldir = re.sub('.+/XML', '', os.path.dirname(v))
            output_dir = xmldir
            #if output_dir[0] == '/':
            #    output_dir = output_dir[1:]
            output_dir = output_dir.strip('/')
            output_dir = os.path.join(output, output_dir)
        generator = Generator(target=target, ignore_errors=ignore_errors, allgenerator=allgenerator)
        generator.functionlist(output_dir, doxygen.rootnodes[index].children)
        generator.functionlist(output_dir, doxygen.groupsnodes[index].children)
        generator.enumlist(output_dir, doxygen.groupsnodes[index].children)
        generator.definelist(output_dir, doxygen.groupsnodes[index].children)
        generator.definelist(output_dir, doxygen.rootnodes[index].children)
        generator.definelist(output_dir, doxygen.filesnodes[index].children)
        generator.functions(output_dir, doxygen.rootnodes[index].children)
        generator.functions(output_dir, doxygen.groupsnodes[index].children)
        generator.enums(output_dir, doxygen.rootnodes[index].children)
        generator.enums(output_dir, doxygen.groupsnodes[index].children)
        generator.annotated(output_dir, doxygen.rootnodes[index].children)
        generator.fileindex(output_dir, doxygen.filesnodes[index].children)
        generator.members(output_dir, doxygen.rootnodes[index].children)
        generator.members(output_dir, doxygen.groupsnodes[index].children)
        generator.files(output_dir, doxygen.filesnodes[index].children)
        generator.namespaces(output_dir, doxygen.rootnodes[index].children)
        generator.classes(output_dir, doxygen.rootnodes[index].children)
        generator.hierarchy(output_dir, doxygen.rootnodes[index].children)
        generator.modules(output_dir, doxygen.groupsnodes[index].children)
        generator.pages(output_dir, doxygen.pagesnodes[index].children)
        generator.relatedpages(output_dir, doxygen.pagesnodes[index].children)
        generator.index(output_dir, doxygen.rootnodes[index].children, [Kind.FUNCTION, Kind.VARIABLE, Kind.TYPEDEF, Kind.ENUM], [Kind.CLASS, Kind.STRUCT, Kind.INTERFACE], 'Class Members')
        generator.index(output_dir, doxygen.rootnodes[index].children, [Kind.FUNCTION], [Kind.CLASS, Kind.STRUCT, Kind.INTERFACE], 'Class Member Functions')
        generator.index(output_dir, doxygen.rootnodes[index].children, [Kind.VARIABLE], [Kind.CLASS, Kind.STRUCT, Kind.INTERFACE], 'Class Member Variables')
        generator.index(output_dir, doxygen.rootnodes[index].children, [Kind.TYPEDEF], [Kind.CLASS, Kind.STRUCT, Kind.INTERFACE], 'Class Member Typedefs')
        generator.index(output_dir, doxygen.rootnodes[index].children, [Kind.ENUM], [Kind.CLASS, Kind.STRUCT, Kind.INTERFACE], 'Class Member Enums')
        generator.index(output_dir, doxygen.rootnodes[index].children, [Kind.FUNCTION, Kind.VARIABLE, Kind.TYPEDEF, Kind.ENUM], [Kind.NAMESPACE], 'Namespace Members')
        generator.index(output_dir, doxygen.rootnodes[index].children, [Kind.FUNCTION], [Kind.NAMESPACE], 'Namespace Member Functions')
        generator.index(output_dir, doxygen.rootnodes[index].children, [Kind.VARIABLE], [Kind.NAMESPACE], 'Namespace Member Variables')
        generator.index(output_dir, doxygen.rootnodes[index].children, [Kind.TYPEDEF], [Kind.NAMESPACE], 'Namespace Member Typedefs')
        generator.index(output_dir, doxygen.rootnodes[index].children, [Kind.ENUM], [Kind.NAMESPACE], 'Namespace Member Enums')
        generator.index(output_dir, doxygen.filesnodes[index].children, [Kind.FUNCTION], [Kind.FILE], 'Functions')
        generator.index(output_dir, doxygen.filesnodes[index].children, [Kind.DEFINE], [Kind.FILE], 'Macros')
        generator.index(output_dir, doxygen.filesnodes[index].children, [Kind.VARIABLE, Kind.UNION, Kind.TYPEDEF, Kind.ENUM], [Kind.FILE], 'Variables')

        if target == 'gitbook' and summary:
            generator.summary(output_dir, summary, \
                    doxygen.rootnodes[index].children, doxygen.groupsnodes[index].children, \
                    doxygen.filesnodes[index].children, doxygen.pagesnodes[index].children,\
                    no_example = no_example)
